@@ -33,13 +33,14 @@
 #endif
 
 /* -------------------------------------------------------------------- */
-/* Common								*/
+/* Common                                                               */
 /* -------------------------------------------------------------------- */
 
 typedef struct {
     PyObject_HEAD
     int (*encode)(Imaging im, ImagingCodecState state,
-		  UINT8* buffer, int bytes);
+                  UINT8* buffer, int bytes);
+    int (*cleanup)(ImagingCodecState state);
     struct ImagingCodecStateInstance state;
     Imaging im;
     PyObject* lock;
@@ -58,25 +59,28 @@ PyImaging_EncoderNew(int contextsize)
 
     encoder = PyObject_New(ImagingEncoderObject, &ImagingEncoderType);
     if (encoder == NULL)
-	return NULL;
+        return NULL;
 
     /* Clear the encoder state */
     memset(&encoder->state, 0, sizeof(encoder->state));
 
     /* Allocate encoder context */
     if (contextsize > 0) {
-	context = (void*) calloc(1, contextsize);
-	if (!context) {
-	    Py_DECREF(encoder);
-	    (void) PyErr_NoMemory();
-	    return NULL;
-	}
+        context = (void*) calloc(1, contextsize);
+        if (!context) {
+            Py_DECREF(encoder);
+            (void) PyErr_NoMemory();
+            return NULL;
+        }
     } else
-	context = 0;
+        context = 0;
 
     /* Initialize encoder context */
     encoder->state.context = context;
 
+    /* Most encoders don't need this */
+    encoder->cleanup = NULL;
+
     /* Target image */
     encoder->lock = NULL;
     encoder->im = NULL;
@@ -87,12 +91,26 @@ PyImaging_EncoderNew(int contextsize)
 static void
 _dealloc(ImagingEncoderObject* encoder)
 {
+    if (encoder->cleanup)
+        encoder->cleanup(&encoder->state);
     free(encoder->state.buffer);
     free(encoder->state.context);
     Py_XDECREF(encoder->lock);
     PyObject_Del(encoder);
 }
 
+static PyObject*
+_encode_cleanup(ImagingEncoderObject* encoder, PyObject* args)
+{
+    int status = 0;
+
+    if (encoder->cleanup){
+        status = encoder->cleanup(&encoder->state);
+    }
+
+    return Py_BuildValue("i", status);
+}
+
 static PyObject*
 _encode(ImagingEncoderObject* encoder, PyObject* args)
 {
@@ -105,14 +123,14 @@ _encode(ImagingEncoderObject* encoder, PyObject* args)
     int bufsize = 16384;
 
     if (!PyArg_ParseTuple(args, "|i", &bufsize))
-	return NULL;
+        return NULL;
 
     buf = PyBytes_FromStringAndSize(NULL, bufsize);
     if (!buf)
-	return NULL;
+        return NULL;
 
     status = encoder->encode(encoder->im, &encoder->state,
-			     (UINT8*) PyBytes_AsString(buf), bufsize);
+                             (UINT8*) PyBytes_AsString(buf), bufsize);
 
     /* adjust string length to avoid slicing in encoder */
     if (_PyBytes_Resize(&buf, (status > 0) ? status : 0) < 0)
@@ -138,28 +156,28 @@ _encode_to_file(ImagingEncoderObject* encoder, PyObject* args)
     int bufsize = 16384;
 
     if (!PyArg_ParseTuple(args, "i|i", &fh, &bufsize))
-	return NULL;
+        return NULL;
 
     /* Allocate an encoder buffer */
     buf = (UINT8*) malloc(bufsize);
     if (!buf)
-	return PyErr_NoMemory();
+        return PyErr_NoMemory();
 
     ImagingSectionEnter(&cookie);
 
     do {
 
-	/* This replaces the inner loop in the ImageFile _save
-	   function. */
+        /* This replaces the inner loop in the ImageFile _save
+           function. */
 
-	status = encoder->encode(encoder->im, &encoder->state, buf, bufsize);
+        status = encoder->encode(encoder->im, &encoder->state, buf, bufsize);
 
-	if (status > 0)
-	    if (write(fh, buf, status) < 0) {
+        if (status > 0)
+            if (write(fh, buf, status) < 0) {
                 ImagingSectionLeave(&cookie);
-		free(buf);
-		return PyErr_SetFromErrno(PyExc_IOError);
-	    }
+                free(buf);
+                return PyErr_SetFromErrno(PyExc_IOError);
+            }
 
     } while (encoder->state.errcode == 0);
 
@@ -186,39 +204,39 @@ _setimage(ImagingEncoderObject* encoder, PyObject* args)
 
     /* FIXME: should publish the ImagingType descriptor */
     if (!PyArg_ParseTuple(args, "O|(iiii)", &op, &x0, &y0, &x1, &y1))
-	return NULL;
+        return NULL;
     im = PyImaging_AsImaging(op);
     if (!im)
-	return NULL;
+        return NULL;
 
     encoder->im = im;
 
     state = &encoder->state;
 
     if (x0 == 0 && x1 == 0) {
-	state->xsize = im->xsize;
-	state->ysize = im->ysize;
+        state->xsize = im->xsize;
+        state->ysize = im->ysize;
     } else {
-	state->xoff = x0;
-	state->yoff = y0;
-	state->xsize = x1 - x0;
-	state->ysize = y1 - y0;
+        state->xoff = x0;
+        state->yoff = y0;
+        state->xsize = x1 - x0;
+        state->ysize = y1 - y0;
     }
 
     if (state->xsize <= 0 ||
-	state->xsize + state->xoff > im->xsize ||
-	state->ysize <= 0 ||
-	state->ysize + state->yoff > im->ysize) {
-	PyErr_SetString(PyExc_SystemError, "tile cannot extend outside image");
-	return NULL;
+        state->xsize + state->xoff > im->xsize ||
+        state->ysize <= 0 ||
+        state->ysize + state->yoff > im->ysize) {
+        PyErr_SetString(PyExc_SystemError, "tile cannot extend outside image");
+        return NULL;
     }
 
     /* Allocate memory buffer (if bits field is set) */
     if (state->bits > 0) {
-	state->bytes = (state->bits * state->xsize+7)/8;
-	state->buffer = (UINT8*) malloc(state->bytes);
-	if (!state->buffer)
-	    return PyErr_NoMemory();
+        state->bytes = (state->bits * state->xsize+7)/8;
+        state->buffer = (UINT8*) malloc(state->bytes);
+        if (!state->buffer)
+            return PyErr_NoMemory();
     }
 
     /* Keep a reference to the image object, to make sure it doesn't
@@ -233,19 +251,20 @@ _setimage(ImagingEncoderObject* encoder, PyObject* args)
 
 static struct PyMethodDef methods[] = {
     {"encode", (PyCFunction)_encode, 1},
+    {"cleanup", (PyCFunction)_encode_cleanup, 1},
     {"encode_to_file", (PyCFunction)_encode_to_file, 1},
     {"setimage", (PyCFunction)_setimage, 1},
     {NULL, NULL} /* sentinel */
 };
 
 static PyTypeObject ImagingEncoderType = {
-	PyVarObject_HEAD_INIT(NULL, 0)
-	"ImagingEncoder",		/*tp_name*/
-	sizeof(ImagingEncoderObject),	/*tp_size*/
-	0,				/*tp_itemsize*/
-	/* methods */
-	(destructor)_dealloc,		/*tp_dealloc*/
-	0,				/*tp_print*/
+    PyVarObject_HEAD_INIT(NULL, 0)
+    "ImagingEncoder",               /*tp_name*/
+    sizeof(ImagingEncoderObject),   /*tp_size*/
+    0,                              /*tp_itemsize*/
+    /* methods */
+    (destructor)_dealloc,           /*tp_dealloc*/
+    0,                              /*tp_print*/
     0,                          /*tp_getattr*/
     0,                          /*tp_setattr*/
     0,                          /*tp_compare*/
@@ -283,9 +302,9 @@ get_packer(ImagingEncoderObject* encoder, const char* mode,
 
     pack = ImagingFindPacker(mode, rawmode, &bits);
     if (!pack) {
-	Py_DECREF(encoder);
-	PyErr_SetString(PyExc_SystemError, "unknown raw mode");
-	return -1;
+        Py_DECREF(encoder);
+        PyErr_SetString(PyExc_SystemError, "unknown raw mode");
+        return -1;
     }
 
     encoder->state.shuffle = pack;
@@ -296,7 +315,7 @@ get_packer(ImagingEncoderObject* encoder, const char* mode,
 
 
 /* -------------------------------------------------------------------- */
-/* EPS									*/
+/* EPS                                                                  */
 /* -------------------------------------------------------------------- */
 
 PyObject*
@@ -306,7 +325,7 @@ PyImaging_EpsEncoderNew(PyObject* self, PyObject* args)
 
     encoder = PyImaging_EncoderNew(0);
     if (encoder == NULL)
-	return NULL;
+        return NULL;
 
     encoder->encode = ImagingEpsEncode;
 
@@ -315,7 +334,7 @@ PyImaging_EpsEncoderNew(PyObject* self, PyObject* args)
 
 
 /* -------------------------------------------------------------------- */
-/* GIF									*/
+/* GIF                                                                  */
 /* -------------------------------------------------------------------- */
 
 PyObject*
@@ -328,14 +347,14 @@ PyImaging_GifEncoderNew(PyObject* self, PyObject* args)
     int bits = 8;
     int interlace = 0;
     if (!PyArg_ParseTuple(args, "ss|ii", &mode, &rawmode, &bits, &interlace))
-	return NULL;
+        return NULL;
 
     encoder = PyImaging_EncoderNew(sizeof(GIFENCODERSTATE));
     if (encoder == NULL)
-	return NULL;
+        return NULL;
 
     if (get_packer(encoder, mode, rawmode) < 0)
-	return NULL;
+        return NULL;
 
     encoder->encode = ImagingGifEncode;
 
@@ -347,7 +366,7 @@ PyImaging_GifEncoderNew(PyObject* self, PyObject* args)
 
 
 /* -------------------------------------------------------------------- */
-/* PCX									*/
+/* PCX                                                                  */
 /* -------------------------------------------------------------------- */
 
 PyObject*
@@ -358,15 +377,19 @@ PyImaging_PcxEncoderNew(PyObject* self, PyObject* args)
     char *mode;
     char *rawmode;
     int bits = 8;
-    if (!PyArg_ParseTuple(args, "ss|ii", &mode, &rawmode, &bits))
-	return NULL;
+
+    if (!PyArg_ParseTuple(args, "ss|ii", &mode, &rawmode, &bits)) {
+        return NULL;
+    }
 
     encoder = PyImaging_EncoderNew(0);
-    if (encoder == NULL)
-	return NULL;
+    if (encoder == NULL) {
+        return NULL;
+    }
 
-    if (get_packer(encoder, mode, rawmode) < 0)
-	return NULL;
+    if (get_packer(encoder, mode, rawmode) < 0) {
+        return NULL;
+    }
 
     encoder->encode = ImagingPcxEncode;
 
@@ -375,7 +398,7 @@ PyImaging_PcxEncoderNew(PyObject* self, PyObject* args)
 
 
 /* -------------------------------------------------------------------- */
-/* RAW									*/
+/* RAW                                                                  */
 /* -------------------------------------------------------------------- */
 
 PyObject*
@@ -389,14 +412,14 @@ PyImaging_RawEncoderNew(PyObject* self, PyObject* args)
     int ystep = 1;
 
     if (!PyArg_ParseTuple(args, "ss|ii", &mode, &rawmode, &stride, &ystep))
-	return NULL;
+        return NULL;
 
     encoder = PyImaging_EncoderNew(0);
     if (encoder == NULL)
-	return NULL;
+        return NULL;
 
     if (get_packer(encoder, mode, rawmode) < 0)
-	return NULL;
+        return NULL;
 
     encoder->encode = ImagingRawEncode;
 
@@ -408,7 +431,7 @@ PyImaging_RawEncoderNew(PyObject* self, PyObject* args)
 
 
 /* -------------------------------------------------------------------- */
-/* XBM									*/
+/* XBM                                                                  */
 /* -------------------------------------------------------------------- */
 
 PyObject*
@@ -418,10 +441,10 @@ PyImaging_XbmEncoderNew(PyObject* self, PyObject* args)
 
     encoder = PyImaging_EncoderNew(0);
     if (encoder == NULL)
-	return NULL;
+        return NULL;
 
     if (get_packer(encoder, "1", "1;R") < 0)
-	return NULL;
+        return NULL;
 
     encoder->encode = ImagingXbmEncode;
 
@@ -430,7 +453,7 @@ PyImaging_XbmEncoderNew(PyObject* self, PyObject* args)
 
 
 /* -------------------------------------------------------------------- */
-/* ZIP									*/
+/* ZIP                                                                  */
 /* -------------------------------------------------------------------- */
 
 #ifdef HAVE_LIBZ
@@ -469,16 +492,16 @@ PyImaging_ZipEncoderNew(PyObject* self, PyObject* args)
 
     encoder = PyImaging_EncoderNew(sizeof(ZIPSTATE));
     if (encoder == NULL)
-	return NULL;
+        return NULL;
 
     if (get_packer(encoder, mode, rawmode) < 0)
-	return NULL;
+        return NULL;
 
     encoder->encode = ImagingZipEncode;
 
     if (rawmode[0] == 'P')
-	/* disable filtering */
-	((ZIPSTATE*)encoder->state.context)->mode = ZIP_PNG_PALETTE;
+        /* disable filtering */
+        ((ZIPSTATE*)encoder->state.context)->mode = ZIP_PNG_PALETTE;
 
     ((ZIPSTATE*)encoder->state.context)->optimize = optimize;
     ((ZIPSTATE*)encoder->state.context)->compress_level = compress_level;
@@ -492,7 +515,7 @@ PyImaging_ZipEncoderNew(PyObject* self, PyObject* args)
 
 
 /* -------------------------------------------------------------------- */
-/* JPEG									*/
+/* JPEG                                                                 */
 /* -------------------------------------------------------------------- */
 
 #ifdef HAVE_LIBJPEG
@@ -500,24 +523,24 @@ PyImaging_ZipEncoderNew(PyObject* self, PyObject* args)
 /* We better define this encoder last in this file, so the following
    undef's won't mess things up for the Imaging library proper. */
 
-#undef	HAVE_PROTOTYPES
-#undef	HAVE_STDDEF_H
-#undef	HAVE_STDLIB_H
-#undef	UINT8
-#undef	UINT16
-#undef	UINT32
-#undef	INT8
-#undef	INT16
-#undef	INT32
+#undef  HAVE_PROTOTYPES
+#undef  HAVE_STDDEF_H
+#undef  HAVE_STDLIB_H
+#undef  UINT8
+#undef  UINT16
+#undef  UINT32
+#undef  INT8
+#undef  INT16
+#undef  INT32
 
 #include "Jpeg.h"
 
-static unsigned int** get_qtables_arrays(PyObject* qtables) {
+static unsigned int* get_qtables_arrays(PyObject* qtables, int* qtablesLen) {
     PyObject* tables;
     PyObject* table;
     PyObject* table_data;
     int i, j, num_tables;
-    unsigned int **qarrays;
+    unsigned int *qarrays;
 
     if ((qtables ==  NULL) || (qtables == Py_None)) {
         return NULL;
@@ -530,11 +553,13 @@ static unsigned int** get_qtables_arrays(PyObject* qtables) {
 
     tables = PySequence_Fast(qtables, "expected a sequence");
     num_tables = PySequence_Size(qtables);
-    if (num_tables < 2 || num_tables > NUM_QUANT_TBLS) {
-        PyErr_SetString(PyExc_ValueError, "Not a valid numbers of quantization tables. Should be between 2 and 4.");
+    if (num_tables < 1 || num_tables > NUM_QUANT_TBLS) {
+        PyErr_SetString(PyExc_ValueError, 
+            "Not a valid number of quantization tables. Should be between 1 and 4.");
+        Py_DECREF(tables);
         return NULL;
     }
-    qarrays = (unsigned int**) PyMem_Malloc(num_tables * sizeof(unsigned int*));
+    qarrays = (unsigned int*) malloc(num_tables * DCTSIZE2 * sizeof(unsigned int));
     if (!qarrays) {
         Py_DECREF(tables);
         PyErr_NoMemory();
@@ -543,38 +568,33 @@ static unsigned int** get_qtables_arrays(PyObject* qtables) {
     for (i = 0; i < num_tables; i++) {
         table = PySequence_Fast_GET_ITEM(tables, i);
         if (!PySequence_Check(table)) {
-            Py_DECREF(tables);
             PyErr_SetString(PyExc_ValueError, "Invalid quantization tables");
-            return NULL;
+            goto JPEG_QTABLES_ERR;
         }
         if (PySequence_Size(table) != DCTSIZE2) {
-            Py_DECREF(tables);
-            PyErr_SetString(PyExc_ValueError, "Invalid quantization tables");
-            return NULL;
+            PyErr_SetString(PyExc_ValueError, "Invalid quantization table size");
+            goto JPEG_QTABLES_ERR;
         }
         table_data = PySequence_Fast(table, "expected a sequence");
-        qarrays[i] = (unsigned int*) PyMem_Malloc(DCTSIZE2 * sizeof(unsigned int));
-        if (!qarrays[i]) {
-            Py_DECREF(tables);
-            PyErr_NoMemory();
-            return NULL;
-        }
         for (j = 0; j < DCTSIZE2; j++) {
-            qarrays[i][j] = PyInt_AS_LONG(PySequence_Fast_GET_ITEM(table_data, j));
+            qarrays[i * DCTSIZE2 + j] = PyInt_AS_LONG(PySequence_Fast_GET_ITEM(table_data, j));
         }
+        Py_DECREF(table_data);
     }
 
-    Py_DECREF(tables);
+    *qtablesLen = num_tables; 
 
+JPEG_QTABLES_ERR:
+    Py_DECREF(tables);  // Run on both error and not error
     if (PyErr_Occurred()) {
-        PyMem_Free(qarrays);
+        free(qarrays);
         qarrays = NULL;
+        return NULL;
     }
 
     return qarrays;
 }
 
-
 PyObject*
 PyImaging_JpegEncoderNew(PyObject* self, PyObject* args)
 {
@@ -590,7 +610,8 @@ PyImaging_JpegEncoderNew(PyObject* self, PyObject* args)
     int xdpi = 0, ydpi = 0;
     int subsampling = -1; /* -1=default, 0=none, 1=medium, 2=high */
     PyObject* qtables=NULL;
-    unsigned int **qarrays = NULL;
+    unsigned int *qarrays = NULL;
+    int qtablesLen = 0;
     char* extra = NULL;
     int extra_size;
     char* rawExif = NULL;
@@ -601,19 +622,19 @@ PyImaging_JpegEncoderNew(PyObject* self, PyObject* args)
                           &progressive, &smooth, &optimize, &streamtype,
                           &xdpi, &ydpi, &subsampling, &qtables, &extra, &extra_size,
                           &rawExif, &rawExifLen))
-	return NULL;
+        return NULL;
 
     encoder = PyImaging_EncoderNew(sizeof(JPEGENCODERSTATE));
     if (encoder == NULL)
-	return NULL;
+        return NULL;
 
     if (get_packer(encoder, mode, rawmode) < 0)
-	return NULL;
+        return NULL;
 
-    qarrays = get_qtables_arrays(qtables);
+    qarrays = get_qtables_arrays(qtables, &qtablesLen);
 
     if (extra && extra_size > 0) {
-        char* p = malloc(extra_size);
+        char* p = malloc(extra_size); // Freed in JpegEncode, Case 5
         if (!p)
             return PyErr_NoMemory();
         memcpy(p, extra, extra_size);
@@ -622,7 +643,7 @@ PyImaging_JpegEncoderNew(PyObject* self, PyObject* args)
         extra = NULL;
 
     if (rawExif && rawExifLen > 0) {
-        char* pp = malloc(rawExifLen);
+        char* pp = malloc(rawExifLen); // Freed in JpegEncode, Case 5
         if (!pp)
             return PyErr_NoMemory();
         memcpy(pp, rawExif, rawExifLen);
@@ -634,6 +655,7 @@ PyImaging_JpegEncoderNew(PyObject* self, PyObject* args)
 
     ((JPEGENCODERSTATE*)encoder->state.context)->quality = quality;
     ((JPEGENCODERSTATE*)encoder->state.context)->qtables = qarrays;
+    ((JPEGENCODERSTATE*)encoder->state.context)->qtablesLen = qtablesLen;
     ((JPEGENCODERSTATE*)encoder->state.context)->subsampling = subsampling;
     ((JPEGENCODERSTATE*)encoder->state.context)->progressive = progressive;
     ((JPEGENCODERSTATE*)encoder->state.context)->smooth = smooth;
@@ -660,9 +682,6 @@ PyImaging_JpegEncoderNew(PyObject* self, PyObject* args)
 #include "TiffDecode.h"
 
 #include <string.h>
-#ifdef __WIN32__
-#define strcasecmp(s1, s2) stricmp(s1, s2)
-#endif
 
 PyObject*
 PyImaging_LibTiffEncoderNew(PyObject* self, PyObject* args)
@@ -718,11 +737,11 @@ PyImaging_LibTiffEncoderNew(PyObject* self, PyObject* args)
         return NULL;
     }
 
-	// While failes on 64 bit machines, complains that pos is an int instead of a Py_ssize_t
-	//    while (PyDict_Next(dir, &pos, &key, &value)) {
-	for (pos=0;pos<d_size;pos++){
-		key = PyList_GetItem(keys,pos);
-		value = PyList_GetItem(values,pos);
+        // While failes on 64 bit machines, complains that pos is an int instead of a Py_ssize_t
+        //    while (PyDict_Next(dir, &pos, &key, &value)) {
+        for (pos=0;pos<d_size;pos++){
+                key = PyList_GetItem(keys,pos);
+                value = PyList_GetItem(values,pos);
         status = 0;
         TRACE(("Attempting to set key: %d\n", (int)PyInt_AsLong(key)));
         if (PyInt_Check(value)) {
@@ -753,7 +772,7 @@ PyImaging_LibTiffEncoderNew(PyObject* self, PyObject* args)
                                                         (ttag_t) PyInt_AsLong(key),
                                                         intav);
                         free(intav);
-                    }       
+                    }
                 } else {
                     TRACE((" %d elements, setting as floats \n", len));
                     floatav = malloc(sizeof(float)*len);
@@ -793,3 +812,158 @@ PyImaging_LibTiffEncoderNew(PyObject* self, PyObject* args)
 
 #endif
 
+/* -------------------------------------------------------------------- */
+/* JPEG	2000								*/
+/* -------------------------------------------------------------------- */
+
+#ifdef HAVE_OPENJPEG
+
+#include "Jpeg2K.h"
+
+static void
+j2k_decode_coord_tuple(PyObject *tuple, int *x, int *y)
+{
+    *x = *y = 0;
+
+    if (tuple && PyTuple_Check(tuple) && PyTuple_GET_SIZE(tuple) == 2) {
+        *x = (int)PyInt_AsLong(PyTuple_GET_ITEM(tuple, 0));
+        *y = (int)PyInt_AsLong(PyTuple_GET_ITEM(tuple, 1));
+
+        if (*x < 0)
+            *x = 0;
+        if (*y < 0)
+            *y = 0;
+    }
+}
+
+PyObject*
+PyImaging_Jpeg2KEncoderNew(PyObject *self, PyObject *args)
+{
+    ImagingEncoderObject *encoder;
+    JPEG2KENCODESTATE *context;
+
+    char *mode;
+    char *format;
+    OPJ_CODEC_FORMAT codec_format;
+    PyObject *offset = NULL, *tile_offset = NULL, *tile_size = NULL;
+    char *quality_mode = "rates";
+    PyObject *quality_layers = NULL;
+    int num_resolutions = 0;
+    PyObject *cblk_size = NULL, *precinct_size = NULL;
+    PyObject *irreversible = NULL;
+    char *progression = "LRCP";
+    OPJ_PROG_ORDER prog_order;
+    char *cinema_mode = "no";
+    OPJ_CINEMA_MODE cine_mode;
+    int fd = -1;
+
+    if (!PyArg_ParseTuple(args, "ss|OOOsOIOOOssi", &mode, &format,
+                          &offset, &tile_offset, &tile_size,
+                          &quality_mode, &quality_layers, &num_resolutions,
+                          &cblk_size, &precinct_size,
+                          &irreversible, &progression, &cinema_mode,
+                          &fd))
+        return NULL;
+
+    if (strcmp (format, "j2k") == 0)
+        codec_format = OPJ_CODEC_J2K;
+    else if (strcmp (format, "jpt") == 0)
+        codec_format = OPJ_CODEC_JPT;
+    else if (strcmp (format, "jp2") == 0)
+        codec_format = OPJ_CODEC_JP2;
+    else
+        return NULL;
+
+    if (strcmp(progression, "LRCP") == 0)
+        prog_order = OPJ_LRCP;
+    else if (strcmp(progression, "RLCP") == 0)
+        prog_order = OPJ_RLCP;
+    else if (strcmp(progression, "RPCL") == 0)
+        prog_order = OPJ_RPCL;
+    else if (strcmp(progression, "PCRL") == 0)
+        prog_order = OPJ_PCRL;
+    else if (strcmp(progression, "CPRL") == 0)
+        prog_order = OPJ_CPRL;
+    else
+        return NULL;
+
+    if (strcmp(cinema_mode, "no") == 0)
+        cine_mode = OPJ_OFF;
+    else if (strcmp(cinema_mode, "cinema2k-24") == 0)
+        cine_mode = OPJ_CINEMA2K_24;
+    else if (strcmp(cinema_mode, "cinema2k-48") == 0)
+        cine_mode = OPJ_CINEMA2K_48;
+    else if (strcmp(cinema_mode, "cinema4k-24") == 0)
+        cine_mode = OPJ_CINEMA4K_24;
+    else
+        return NULL;
+
+    encoder = PyImaging_EncoderNew(sizeof(JPEG2KENCODESTATE));
+    if (!encoder)
+        return NULL;
+
+    encoder->encode = ImagingJpeg2KEncode;
+    encoder->cleanup = ImagingJpeg2KEncodeCleanup;
+
+    context = (JPEG2KENCODESTATE *)encoder->state.context;
+
+    context->fd = fd;
+    context->format = codec_format;
+    context->offset_x = context->offset_y = 0;
+
+    j2k_decode_coord_tuple(offset, &context->offset_x, &context->offset_y);
+    j2k_decode_coord_tuple(tile_offset,
+                           &context->tile_offset_x,
+                           &context->tile_offset_y);
+    j2k_decode_coord_tuple(tile_size,
+                           &context->tile_size_x,
+                           &context->tile_size_y);
+
+    /* Error on illegal tile offsets */
+    if (context->tile_size_x && context->tile_size_y) {
+        if (context->tile_offset_x <= context->offset_x - context->tile_size_x
+            || context->tile_offset_y <= context->offset_y - context->tile_size_y) {
+            PyErr_SetString(PyExc_ValueError,
+                            "JPEG 2000 tile offset too small; top left tile must "
+                            "intersect image area");
+        }
+
+        if (context->tile_offset_x > context->offset_x
+            || context->tile_offset_y > context->offset_y) {
+            PyErr_SetString(PyExc_ValueError,
+                            "JPEG 2000 tile offset too large to cover image area");
+            Py_DECREF(encoder);
+            return NULL;
+        }
+    }
+
+    if (quality_layers && PySequence_Check(quality_layers)) {
+        context->quality_is_in_db = strcmp (quality_mode, "dB") == 0;
+        context->quality_layers = quality_layers;
+        Py_INCREF(quality_layers);
+    }
+
+    context->num_resolutions = num_resolutions;
+
+    j2k_decode_coord_tuple(cblk_size,
+                           &context->cblk_width,
+                           &context->cblk_height);
+    j2k_decode_coord_tuple(precinct_size,
+                           &context->precinct_width,
+                           &context->precinct_height);
+
+    context->irreversible = PyObject_IsTrue(irreversible);
+    context->progression = prog_order;
+    context->cinema_mode = cine_mode;
+
+    return (PyObject *)encoder;
+}
+
+#endif
+
+/*
+ * Local Variables:
+ * c-basic-offset: 4
+ * End:
+ *
+ */
